// Generated by CoffeeScript 1.4.0

/*
The general structure of an incrementally update-able visualization follows these steps:

1. Gather the parameters you'll need to specify the visualization
   a. Gather some info from Rally's standard WSAPI
   b. Gather some info from the user.

2. Create a hash from info from above to be used as the key for cache lookup.

3. Restore the cached calculation using LocalCache.

4. Render the cached calculation. Leave space for updates on the x-axis. Show spinners for missing parts.

5. Query the Lookback API for the incremental "snapshots" not found in the cache.
   Get one page's worth of updates. Maybe 10,000 snapshots max?

6. Update the calculation/manipulation/aggregation of the snapshot data.

7. Update the chart.

8. If there are still more pages of snapshots to update repeat starting at step 5.
*/


(function() {
  var Time, VisualizerBase, lumenize, utils,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; },
    __slice = [].slice;

  if (typeof exports !== "undefined" && exports !== null) {
    lumenize = require('../lib/Lumenize');
  } else {
    lumenize = require('/lumenize');
  }

  utils = lumenize.utils, Time = lumenize.Time;

  VisualizerBase = (function() {
    /*
      @class ChartVisualizerBase
        This is intended to the be the base class for ChartVisualizers. It assumes a template method pattern where the parts
        of the algorithm that have to do with saving to and restoring from the LocalCache (using localStorage API) and
        providing events for config changes or data updates.
    
        You must override these methods:
          * initialize() - set @LumenizeCalculatorClass (implements Lumenize.iCalculator)
    
        You may wish to override:
          * deriveFields(snapshots)
    
      @cfg {Number} [refreshIntervalMilliseconds = 30 * 60 * 1000] Defaults to 30 minutes
    
      @property {Object} userConfig This is whatever the users passes in under the @userConfig parameter in the constructor. It is useful for creating the cache hash. The contents of this will be visualizer specific
    
      @property {Object} config Starts with all the values in userConfig but more may be added
      @property {Number} [config.refreshIntervalMilliseconds = 5 * 60 * 1000] The chart will automatically refresh after this many milliseconds
      @property {Object} [config.deriveFieldsConfig] If you include this, it will pass it into Lumenize.deriveFields as the config Object every time it gets new snapshots to process.
      @property {Boolean} [config.debug = false]
      @property {Object} config.lumenizeCalculatorConfig The config that will be passed to the Lumenize calculator upon instantiation. Do not put x-axis range info in here.
    
      @property {Object} projectAndWorkspaceScope
      @property {Number} projectAndWorkspaceScope.workspaceOID
      @property {Boolean} projectAndWorkspaceScope.projectScopingUp
      @property {Boolean} projectAndWorkspaceScope.projectScopingDown
      @property {Number} projectAndWorkspaceScope.projectOID
    
      @property {Object} workspaceConfiguration Has whatever fields come from Rally but WorkDays and TimeZone (note Caps) are often used by calculators
    
      @property {Lumenize.iCalculator} LumenizeCalculatorClass Must be set; typically in your initialize() method
    
      @property {Object} visualizationData This is where you store the data that you want to communicate to your visualizations.
        It will be passed into createVisualizationCB.
    
      @property {iAnalyticsQuery} analyticsQuery Instantiate this in your onNewDataAvailable() method.
    
      @property {String} upToDateISOString A ISOString (e.g. '2012-01-01T12:34:56.789Z') indicating the last moment that this chart is
        up to date. You should not set this but you can read from it. It will be set when new snapshots are added or it's
        restored from the cache.
      @readonly
    */

    /*
      Sequence diagram below can be edited here: http://www.asciiflow.com/#Draw2041780197906655348/1887977824
     +----------------------+ +-----------------------+ +---------------------+ +--------------------+ +-------------------+ +-----------------------+ +-------------------+ +---------------------+ +---------------+
     |initialize and before | |onConfigOrScopeUpdated | |createVisualization  | |onNewDataAvailable  | |onSnapshotsReceived| |deriveFieldsOnSnapshots| |updateCalculator   | |updateVisualization  | |newDataExpected|
     |----------------------| |-----------------------| |---------------------| |--------------------| |-------------------| |-----------------------| |-------------------| |---------------------| |---------------|
     |@userConfig           | |@lumenizeCalculator    | |@visualizationData   | |@upToDateISOString  | |@upToDateISOString | |snapshots              | |@lumenizeCalculator| |@visualizationData   | |               |
     |@config               | |@upToDateISOString     | | via call to         | | = '2011-12-01...'  | | = endBefore       | |                       | |@cache             | | via call to         | |               |
     |@cache                | | (null if not restored)| | @updateVisualizatio-| | if null            | |@fetchPending      | |                       | |                   | | @updateVisualizatio-| |               |
     |@createVisualizationCB| |@fetchPending = true   | | nData()             | |@analyticsQuery     | |                   | |                       | |                   | | nData()             | |               |
     |                      | |                       | |                     | |@fetchPending = true| |                   | |                       | |                   | |                     | |               |
     +----------------------+ +-----------------------+ +---------------------+ +--------------------+ +-------------------+ +-----------------------+ +-------------------+ +---------------------+ +---------------+
             |                            |                           |                  |                        |                       |                     |                        |                    |
             +--------------------------->|                           |                  |                        |                       |                     |                        |                    |
             |                            +-------------------------->|                  |                        |                       |                     |                        |                    |
             |                            |                           +----------------->|                        |                       |                     |                        |                    |
             |                            |                           |                  +----------------------->|                       |                     |                        |                    |
             |                            |                           |                  |                        +---------------------->|                     |                        |                    |
             |                            |                           |                  |                        |                       +-------------------->|                        |                    |
             |                            |                           |                  |                        |                       |                     +----------------------->|                    |
             |                            |                           |                  |                        |                       |                     |                        +------------------->|
             |                            |                           |                  |                        |                       |                     |                        |                    |
             |                            |                           |                  |<------------------------------- @timeoutHandle = setTimeout(@onNewDataAvailable, delay) ---------------------------+<-+
             |                            |                           |                  |                        |                       |                     |                        |                    |  |
             |                            |                           |                  +----------------------->|                       |                     |                        |                    |  |
             |                            |                           |                  |                        +---------------------->|                     |                        |                    |  |
             |                            |                           |                  |                        |                       +-------------------->|                        |                    |  |
             |                            |                           |                  |                        |                       |                     +----------------------->|                    |  |
             |                            |                           |                  |                        |                       |                     |                        +------------------->+--+
             |                            |                           |                  |                        |                       |                     |                        |                    |
    */

    function VisualizerBase(visualizations, userConfig, createVisualizationCB) {
      this.visualizations = visualizations;
      this.userConfig = userConfig;
      this.createVisualizationCB = createVisualizationCB;
      this.onSnapshotsReceieved = __bind(this.onSnapshotsReceieved, this);

      /*
          You should not have a constructor for the sub-class. Rather, put your code in initialize(). If for some crazy
          reason you really want a constructor, make sure it looks like this:
          ```
          constructor: (myCustomArgument, remainingArguments...) ->
            # Any code you want to execute before initialize(). Use myCustomArgument.
            super(remainingArguments...)
            # Any code you want to execute after initialize(). Use myCustomArgument.
          ```
      */

      this.config = utils.clone(this.userConfig);
      if (this.config.trace) {
        console.log('in VisualizerBase.constructor');
      }
      this.cache = new LocalCache();
      if (this.config.debug == null) {
        this.config.debug = false;
      }
      this.getProjectAndWorkspaceScope();
    }

    VisualizerBase.prototype.getProjectAndWorkspaceScope = function() {
      var projectOID, projectOIDsInScope, projectScopingDown, projectScopingUp, scope, workspaceOID, _callback,
        _this = this;
      if (this.config.trace) {
        console.log('in VisualizerBase.getProjectAndWorkspaceScope');
      }
      if (top === self) {
        workspaceOID = 41529001;
        projectScopingUp = false;
        projectScopingDown = true;
        projectOID = 279050021;
        projectOIDsInScope = [projectOID];
      } else {
        workspaceOID = __WORKSPACE_OID__;
        projectScopingUp = __PROJECT_SCOPING_UP__;
        projectScopingDown = __PROJECT_SCOPING_DOWN__;
        projectOID = __PROJECT_OID__;
        projectOIDsInScope = [__PROJECT_OIDS_IN_SCOPE__];
      }
      scope = {
        workspaceOID: workspaceOID,
        projectScopingUp: projectScopingUp,
        projectScopingDown: projectScopingDown,
        projectOID: projectOID,
        projectOIDsInScope: projectOIDsInScope
      };
      _callback = function(projectAndWorkspaceScope) {
        _this.projectAndWorkspaceScope = projectAndWorkspaceScope;
        return _this.getWorkspaceConfiguration();
      };
      return _callback(scope);
    };

    VisualizerBase.prototype.getWorkspaceConfiguration = function() {
      var workspaceConfiguration, _callback,
        _this = this;
      if (this.config.trace) {
        console.log('in VisualizerBase.getWorkspaceConfiguration');
      }
      workspaceConfiguration = {
        DateFormat: 'MM/dd/yyyy',
        DateTimeFormat: 'MM/dd/yyyy hh:mm:ss a',
        IterationEstimateUnitName: 'Points',
        ReleaseEstimateUnitName: 'Points',
        TaskUnitName: 'Hours',
        TimeTrackerEnabled: true,
        TimeZone: 'America/Denver',
        WorkDays: 'Monday,Tuesday,Wednesday,Thursday,Friday'
      };
      _callback = function(workspaceConfiguration) {
        _this.workspaceConfiguration = workspaceConfiguration;
        _this.initialize();
        return _this.onConfigOrScopeUpdated();
      };
      return _callback(workspaceConfiguration);
    };

    VisualizerBase.prototype.onConfigOrScopeUpdated = function() {
      var savedState;
      if (this.config.trace) {
        console.log('in VisualizerBase.onConfigOrScopeUpdated');
      }
      savedState = void 0;
      if (savedState != null) {
        if (this.config.debug) {
          console.log('Found a saved state in cache. Restoring from savedState. Size:', JSON.stringify(savedState).length);
          console.log(savedState);
        }
        this.lumenizeCalculator = this.LumenizeCalculatorClass.newFromSavedState(savedState);
        this.upToDateISOString = this.lumenizeCalculator.upToDateISOString;
      } else {
        if (this.config.debug) {
          console.log('Did not find a saved state in cache. Calculating from scratch.');
        }
        this.lumenizeCalculator = new this.LumenizeCalculatorClass(this.config.lumenizeCalculatorConfig);
        this.upToDateISOString = null;
      }
      this.fetchPending = true;
      this.createVisualization();
      this.dirty = false;
      return this.onNewDataAvailable();
    };

    VisualizerBase.prototype.getCurrentState = function() {
      var fields, queryConfig, scopeValue, _callback, _ref,
        _this = this;
      if (this.config.trace) {
        console.log('in VisualizerBase.getCurrentState');
      }
      _callback = function(queryHandle) {
        var r;
        console.log(queryHandle);
        _this.currentState = queryHandle.allResults;
        _this.currentObjectIDs = (function() {
          var _i, _len, _ref, _results;
          _ref = this.currentState;
          _results = [];
          for (_i = 0, _len = _ref.length; _i < _len; _i++) {
            r = _ref[_i];
            _results.push(r.ObjectID);
          }
          return _results;
        }).call(_this);
        return _this.onNewDataAvailable();
      };
      queryConfig = {
        'X-RallyIntegrationName': 'Burn Chart (prototype)',
        'X-RallyIntegrationVendor': 'Rally Red Pill',
        'X-RallyIntegrationVersion': '0.2.0',
        workspaceOID: this.projectAndWorkspaceScope.workspaceOID
      };
      this.analyticsQuery = new GuidedAnalyticsQuery(queryConfig);
      if (this.config.scopeValue === 'scope') {
        if (this.projectAndWorkspaceScope.projectScopingUp) {
          if (this.config.debug) {
            console.log('Project scoping up. OIDs in scope: ', this.projectAndWorkspaceScope.projectOIDsInScope);
          }
          this.analyticsQuery.scope('Project', this.projectAndWorkspaceScope.projectOIDsInScope);
        } else if (this.projectAndWorkspaceScope.projectScopingDown) {
          if (this.config.debug) {
            console.log('Project scoping down. Setting _ProjectHierarchy to: ', this.projectAndWorkspaceScope.projectOID);
          }
          this.analyticsQuery.scope('_ProjectHierarchy', this.projectAndWorkspaceScope.projectOID);
        } else {
          if (this.config.debug) {
            console.log('Project with no up or down scoping. Setting Project to: ', this.projectAndWorkspaceScope.projectOID);
          }
          this.analyticsQuery.scope('Project', this.projectAndWorkspaceScope.projectOID);
        }
      } else if (((_ref = this.config.scopeData) != null ? _ref.ObjectID : void 0) != null) {
        scopeValue = this.config.scopeData.ObjectID;
        this.analyticsQuery.scope(this.config.scopeField, scopeValue);
      } else {
        scopeValue = this.config.scopeValue;
        this.analyticsQuery.scope(this.config.scopeField, scopeValue);
      }
      fields = ["ObjectID"];
      this.analyticsQuery.fields(fields);
      if (this.config.leafOnly) {
        this.analyticsQuery.leafOnly();
      }
      if (this.config.type != null) {
        this.analyticsQuery.type(this.config.type);
      }
      this.analyticsQuery.additionalCriteria(this.config.currentStatePredicate);
      this.analyticsQuery.additionalCriteria({
        __At: "current"
      });
      if (this.config.debug) {
        this.analyticsQuery.debug();
        console.log('Requesting current state data ...');
      }
      return this.analyticsQuery.getAll(_callback);
    };

    VisualizerBase.prototype.getAsOfISOString = function() {
      if (this.config.asOf != null) {
        return this.asOfISOString = new Time(this.config.asOf, 'millisecond').getISOStringInTZ(this.config.lumenizeCalculatorConfig.tz);
      } else {
        return this.asOfISOString = Time.getISOStringFromJSDate();
      }
    };

    VisualizerBase.prototype.onSnapshotsReceieved = function(snapshots, startOn, endBefore, queryInstance) {
      var asOfISOString;
      if (queryInstance == null) {
        queryInstance = null;
      }
      if (this.config.trace) {
        console.log('in VisualizerBase.onSnapshotsReceieved');
      }
      if (snapshots.length > 0 && (new Time(endBefore, Time.MILLISECOND, this.config.tz).getJSDate('GMT').getTime() - new Time(startOn, Time.MILLISECOND, this.config.tz).getJSDate('GMT').getTime()) > 5 * 60 * 1000) {
        this.dirty = true;
      } else {
        this.dirty = false;
      }
      this.upToDateISOString = endBefore;
      this.deriveFieldsOnSnapshots(snapshots);
      asOfISOString = this.getAsOfISOString();
      if (asOfISOString < endBefore) {
        endBefore = asOfISOString;
      }
      this.updateCalculator(snapshots, startOn, endBefore);
      if ((this.config.asOf != null) && this.upToDateISOString < this.config.asOf) {
        this.fetchPending = false;
      } else {
        if (this.analyticsQuery.hasMorePages()) {
          this.fetchPending = true;
        } else {
          this.fetchPending = false;
        }
      }
      this.updateVisualization();
      if (!((this.config.asOf != null) && this.upToDateISOString < this.config.asOf)) {
        if (this.analyticsQuery.hasMorePages()) {
          return this.onNewDataAvailable();
        } else {
          return this.newDataExpected(void 0, this.config.refreshIntervalMilliseconds);
        }
      }
    };

    VisualizerBase.prototype.newDataExpected = function(paddingDelay, etlDelay) {
      var delay;
      if (paddingDelay == null) {
        paddingDelay = 30 * 1000;
      }
      if (etlDelay == null) {
        etlDelay = 30 * 60 * 1000;
      }
      if (this.config.trace) {
        console.log('in VisualizerBase.newDataExpected');
      }
      delay = etlDelay + paddingDelay;
      if (this.timeoutHandle != null) {
        clearTimeout(this.timeoutHandle);
      }
      return this.timeoutHandle = setTimeout(this.onConfigOrScopeUpdated, delay);
    };

    VisualizerBase.prototype.removeFromCacheAndRecalculate = function() {
      if (this.config.trace) {
        console.log('in VisualizerBase.removeFromCacheAndRecalculate');
      }
      this.upToDateISOString = null;
      this.cache.removeItem(this.getHashForCache());
      return this.onConfigOrScopeUpdated();
    };

    VisualizerBase.prototype.updateCalculator = function() {
      var endBefore, rest, savedState, snapshots, startOn, _ref;
      snapshots = arguments[0], startOn = arguments[1], endBefore = arguments[2], rest = 4 <= arguments.length ? __slice.call(arguments, 3) : [];
      /*
          @method updateCalculator
            Allows you to incrementally add snapshots to this calculator. It will also update the cache.
          @param {Object[]} snapshots An array of temporal data model snapshots.
          @param {String} startOn A ISOString (e.g. '2012-01-01T12:34:56.789Z') indicating the time start of the period of
            interest. On the second through nth call, this must equal the previous endBefore.
          @param {String} endBefore A ISOString (e.g. '2012-01-01T12:34:56.789Z') indicating the moment just past the time
            period of interest. This should be the ETLDate from the results of your query to the Lookback API.
      */

      if (this.config.trace) {
        console.log('in VisualizerBase.updateCalculator');
      }
      (_ref = this.lumenizeCalculator).addSnapshots.apply(_ref, [snapshots, startOn, endBefore].concat(__slice.call(rest)));
      savedState = this.lumenizeCalculator.getStateForSaving();
      return this.cache.setItem(this.getHashForCache(), savedState);
    };

    VisualizerBase.prototype.initialize = function() {
      this.dirty = true;
      this.virgin = true;
      if (this.config.trace) {
        console.log('in VisualizerBase.initialize');
      }
      if (this.config.lumenizeCalculatorConfig == null) {
        this.config.lumenizeCalculatorConfig = {};
      }
      this.config.lumenizeCalculatorConfig.workDays = this.workspaceConfiguration.WorkDays;
      if (this.userConfig.tz != null) {
        return this.config.lumenizeCalculatorConfig.tz = this.userConfig.tz;
      } else {
        this.config.tz = this.workspaceConfiguration.TimeZone;
        return this.config.lumenizeCalculatorConfig.tz = this.workspaceConfiguration.TimeZone;
      }
    };

    VisualizerBase.prototype.deriveFieldsOnSnapshots = function(snapshots) {
      if (this.config.trace) {
        console.log('in VisualizerBase.deriveFieldsOnSnapshots');
      }
      if (this.config.deriveFieldsOnSnapshotsConfig != null) {
        return Lumenize.deriveFields(snapshots, this.config.deriveFieldsOnSnapshotsConfig);
      }
    };

    VisualizerBase.prototype.createVisualization = function() {
      if (this.config.trace) {
        console.log('in VisualizerBase.createVisualization. @dirty: ', this.dirty);
      }
      this.updateVisualizationData();
      return this.createVisualizationCB(this.visualizationData);
    };

    VisualizerBase.prototype.updateVisualization = function() {
      if (this.config.trace) {
        console.log('in VisualizerBase.updateVisualization. @dirty: ', this.dirty);
      }
      this.updateVisualizationData();
      if (this.dirty || this.virgin) {
        this.dirty = false;
        this.virgin = false;
        return this.createVisualizationCB(this.visualizationData);
      }
    };

    VisualizerBase.prototype.onNewDataAvailable = function() {
      if (this.config.trace) {
        console.log('in VisualizerBase.onNewDataAvailable');
      }
      this.fetchPending = true;
      return this.analyticsQuery.getPage(this.onSnapshotsReceieved);
    };

    VisualizerBase.prototype.updateVisualizationData = function() {
      if (this.config.trace) {
        return console.log('in VisualizerBase.updateVisualizationData');
      }
    };

    VisualizerBase.prototype.getHashForCache = function() {
      if (this.config.trace) {
        return console.log('in VisualizerBase.getHashForCache');
      }
    };

    return VisualizerBase;

  })();

  this.VisualizerBase = VisualizerBase;

}).call(this);
